
from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.text_format import  Parse , MessageToString
from google.protobuf.json_format import MessageToDict , MessageToJson , ParseDict
import os , configparser , re , json , git 
from DefoldSdk.Engine.Settings import Settings
from DefoldSdk.Engine.GameProject import configparser_to_object
from DefoldSdk.Sdk import Displatcher

sdk =  Displatcher.CreateSdk()



class BaseProject : 
    settings = Settings
    def __init__(self,project  = None) : 
        if project is not None : 
            self.project = os.path.abspath(project)
            self.main_folder = os.path.join(self.project,"main")
            self.gameproject = configparser.ConfigParser()
            self.gameproject.read(os.path.join(self.project,"game.project"))
            self.gameproject = configparser_to_object(self.gameproject)
            self.title = self.gameproject.project.title
            self.files = dict()
            self.default_action()

        




    def default_action(self) : 
        self.MaterialsFolder = os.path.join(self.project,"Materials")
        self.TexturesFolder = os.path.join(self.project,"Textures")
        self.MeshesFolder = os.path.join(self.project,"Meshes")

    def get_relative_path(self,path) : 
        _prj = os.path.abspath(self.project)
        _path = os.path.abspath(path)
        return os.path.relpath(_path,_prj)
    
    def get_project_path(self,path) : 
        return f"/{self.get_relative_path(path)}"

    def projectpath_2_fullpath(self,path) : 
        normal_path = path.removeprefix("/")
        fullpath = os.path.join(self.project,normal_path)
        #assert os.path.exists(fullpath)
        return fullpath

    def get_saved_as(self,obj) : 
        return os.path.join(self.saving_folder(obj) , self.get_filename_saving(obj))

    def update(self) : 
        self.gameproject.write(file = os.path.join(self.project,"game.project"))
        for name , obj in self.files.items() : 
            filename = self.get_filename_saving(obj)
            save_folder = self.saving_folder(obj)
            file_path = os.path.join( save_folder, filename)
            print("<-Exporting->:\t" , file_path)
            proto = obj.to_proto()
 
            print(
                MessageToString(proto) , file = open(file_path,"w")
            )
    def get_filename_saving(self,obj) : 
        template = obj.__form__
        keys = re.findall(r'{(.*?)}', template)
        filename = template.format(**{ key : getattr(obj,key) for key in keys})
        return filename 
    
    def saving_folder(self,obj) : 
        template = self.settings.SavingPolicy[type(obj).__name__]
        keys = re.findall(r'{(.*?)}', template)
        foldername = template.format(**{ key : getattr(obj,key) for key in keys})
        save_folder =  os.path.join(self.project,foldername)
        os.makedirs(save_folder,exist_ok=True )
        return save_folder

    def get_material_by_name(self,name) : 
        material_path = os.path.join(self.MaterialsFolder , name)
        material_path = os.path.join(material_path, f"{name}.material")
        return self.get_project_path(material_path)


    def get_material_texture_by_name(self,name) : 
        texture_path = os.path.join(self.TexturesFolder, f"{name}.png")
        assert os.path.exists(texture_path)
        return self.get_project_path(texture_path)

    def get_mesh_by_name(self,name) : 
        mesh_path = os.path.join(self.MeshesFolder, f"{name}.dae")
        assert os.path.exists(mesh_path)
        return self.get_project_path(mesh_path)

    def is_in_project(self , path):
        path_checking = Path(path).resolve()
        project = Path(self.project).resolve()
        try: return project in path_checking.parents
        except ValueError: return False




class Project(BaseProject) : 

    def new(self, name , path  = "." ,template = None ) : 
        branch  = template if template is not None else self.settings.defaultTemplate 
        projectfolder = os.path.abspath(os.path.join(path,name))
        git.Repo.clone_from(self.settings.Templates, projectfolder, branch=branch)
        self.__init__(project=projectfolder )





    def newCollection(self,**kwargs) : 
        print(sdk , dir(sdk))
        result = sdk.CollectionDesc(**kwargs)
        result.GAME = self 
        self.files[kwargs.get('name')] = result
        return result

    def newMaterial(self,**kwargs) : 

        result = sdk.MaterialDesc(**kwargs)
        result.GAME = self 
        self.files[kwargs.get('name')] = result 
        return result

    def install_extension(self,branch) : 
        git.Repo.clone_from(self.settings.Templates, os.path.join(self.project,branch), branch=branch)






import configparser 


class GameProject : 
	def on_value_changed(self,name, value) : 
		if name in  ["width" , "height"] : 
			if type(value) == str : 
				setattr(self,name,int(value))

	def __setattr__(self, name, value):
		super().__setattr__(name, value)
		self.on_value_changed(name, value)
	
	def write(self,file) : 
		config_object = configparser.ConfigParser()
		for section , sectionvalue in vars(self).items() :
			config_object.add_section(section)
			for field , fieldvalue  in vars(sectionvalue).items() : 
				config_object.set(section, field, str(fieldvalue))
		stream = open(file,"w")
		config_object.write(stream)
		stream.close()



def configparser_to_object(config) :
	sections = config._sections
	result = GameProject()
	for k , v in sections.items() :
		memeber = GameProject()
		for kk , vv in v.items() : 
			setattr(memeber,kk,vv) 
		setattr(result,k,memeber)
	return result












class CameraDesc : 
	__sdk_patch__ = True 
	__ext__ = ".camera"
	def on_field_changed(self,name = None, value = None) : 
		if self.PARENT is not None : 
			self.PARENT.data = MessageToString(self.to_proto(),as_one_line = False)







import os , tempfile , subprocess

class LuaSourceFile : 
	__mule__ = True
	__ext__ = ".script"
	def __init__(self,id   = None, GAME = None , PARENT = None ) :
		self.GAME  = GAME
		self.PARENT  = PARENT
		self.id = id 
		self._filename = self.GAME.get_project_path(self.GAME.get_saved_as(self))
		self._script  = "\n\nfunction init(self)\n\t-- Add initialization code here\n\t-- Learn more: https://defold.com/manuals/script/\n\t-- Remove this function if not needed\nend\n\nfunction final(self)\n\t-- Add finalization code here\n\t-- Learn more: https://defold.com/manuals/script/\n\t-- Remove this function if not needed\nend\n\nfunction update(self, dt)\n\t-- Add update code here\n\t-- Learn more: https://defold.com/manuals/script/\n\t-- Remove this function if not needed\nend\n\nfunction fixed_update(self, dt)\n\t-- This function is called if 'Fixed Update Frequency' is enabled in the Engine section of game.project\n\t-- Can be coupled with fixed updates of the physics simulation if 'Use Fixed Timestep' is enabled in\n\t-- Physics section of game.project\n\t-- Add update code here\n\t-- Learn more: https://defold.com/manuals/script/\n\t-- Remove this function if not needed\nend\n\nfunction on_message(self, message_id, message, sender)\n\t-- Add message-handling code here\n\t-- Learn more: https://defold.com/manuals/message-passing/\n\t-- Remove this function if not needed\nend\n\nfunction on_input(self, action_id, action)\n\t-- Add input-handling code here. The game object this script is attached to\n\t-- must have acquired input focus:\n\t--\n\t--    msg.post(\".\", \"acquire_input_focus\")\n\t--\n\t-- All mapped input bindings will be received. Mouse and touch input will\n\t-- be received regardless of where on the screen it happened.\n\t-- Learn more: https://defold.com/manuals/input/\n\t-- Remove this function if not needed\nend\n\nfunction on_reload(self)\n\t-- Add reload-handling code here\n\t-- Learn more: https://defold.com/manuals/hot-reload/\n\t-- Remove this function if not needed\nend"
		self.on_field_changed()

	@property
	def script(self) : 
		return self._script

	@script.setter
	def script(self,value) : 
		self._script = self.prettyLua(value)
		if self.GAME is not None : 
			with open(self.GAME.projectpath_2_fullpath(self._filename),"w") as buff : 
				buff.write(self.script)
		self.on_field_changed()	

	@property
	def filename(self) : 
		return self._filename

	@filename.setter
	def filename(self,value) : 
		'''
		if is set the old file deleted and the new one will be created and write the script in it 
		'''
		file_fullpath  = self.GAME.projectpath_2_fullpath(self._filename)
		if os.path.exists(file_fullpath) : 
			self.script = open(file_fullpath).read()
			os.remove(file_fullpath)
		self._filename = value
		if self.GAME is not None : 
			with open(self.GAME.projectpath_2_fullpath(self._filename),"w") as buff : 
				buff.write(self.script)
		self.on_field_changed()

	def read(self,file) : 
		if self.GAME.is_in_project(file): 
			with open(file) as buff : self.source = buff.read()
			self.filename = self.GAME.get_project_path(file) 
			self.id = os.path.basename(self.filename).replace(".script","")
		else : 
			# project path 
			full_path = file 
			with open(full_path) as buff : self.source = buff.read()
			if self.GAME.is_in_project(file) : 
				self.filename = self.GAME.get_project_path(file)
				self.id = os.path.basename(self.filename).replace(".script","")
		self.on_field_changed()

		

	def on_field_changed(self,msg = "") : 
		if self.PARENT is not None : 
			self.PARENT.component = self.filename 

	def prettyLua(self,luacode) : 
		result = luacode
		temp_file = tempfile.NamedTemporaryFile(mode='w', delete=False) 
		temp_file.write(luacode)
		temp_file.flush()
		temp_file_path = temp_file.name
		luaformatter_binary = os.path.join(os.path.split(__file__)[0],"lua-format")
		luaformatter_binary_config = os.path.join(os.path.split(__file__)[0],"lua-format.config")
		cmd = [luaformatter_binary,temp_file_path,"-c" , luaformatter_binary_config ]
		process = subprocess.Popen(
				cmd, stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE
		)
		output = bytes.decode(process.stdout.read())
		error = bytes.decode(process.stderr.read())
		if error == "": 
			print(error)
			result = output
		temp_file.close()
		return result



####################################################################################
#*************************************** Sdk *************************************** 
####################################################################################
from PyDefold import Defold
from google.protobuf.descriptor import FieldDescriptor
from google.protobuf.text_format import  Parse , MessageToString
from google.protobuf.json_format import MessageToDict , MessageToJson , ParseDict
import collections 

class SdkListType(list) : 

	def __init__(self,data = list(),PARENT = None , typ = None , name = None) : 
		super().__init__(data)
		self.PARENT = PARENT
		self.name = name  
		self.types = {typ}

	def add_types(self,*args) : 
		self.types.update(args)


	def append(self,item) : 
		assert item.__class__.__name__ in self.types , f"can not append {item.__class__.__name__ } to List[{self.type}]"
		super().append(item)
		if self.PARENT != None : 
			self.PARENT.on_field_changed(msg = "")

	def extend(self,items) : 
		for itm in items : 
			self.append(item=itm )

class SdkType : 
	__id__ = None
	__sdk__ = None 
	def __init__(self,__init__ = True , **kwargs): 
		self.GAME = kwargs.get('GAME' , None)
		self.PARENT = kwargs.get('PARENT' , None)
		self.__preinit__(**kwargs)
		self.__init_message__(**kwargs)
		self.__postinit__(**kwargs)

	def __setAttr__(self,value , name) : 
		setattr(self,f"_{value}",name)
		self.on_field_changed()

	def __init_message__(self,**kwargs) : 
		self.on_field_changed()
		for name in self.fields() : 
			value = kwargs.get(name,None)
			if value is not None : 
				self.__setAttr__(value =  name , name = value )




	def __getAttr__(self,name) : 
		if not hasattr(self,f'_{name}') : self.__init__field__(name)
		return getattr(self,f'_{name}')

	def __init__field__(self,field_name) : 
		field = self.___naitive___.DESCRIPTOR.fields_by_name.get(field_name)
		if field.type == field.TYPE_MESSAGE and field.label != field.LABEL_REPEATED  : # memeber type message
			memeber_type_string  = field.message_type._concrete_class.__name__ 
			message_type = getattr(self.__class__ , memeber_type_string , None) 
			message_type = getattr(self.__class__.__sdk__,memeber_type_string) if message_type is None else message_type
			assert message_type is not None 
			setattr(self,f'_{field_name}' , message_type() )
		if field.type == field.TYPE_MESSAGE and field.label == field.LABEL_REPEATED  : # memeber type list of message
			setattr(self,f'_{field_name}' , SdkListType(typ=field.message_type._concrete_class.__name__ , PARENT = self , name = field_name ) )
		if field.type != field.TYPE_MESSAGE and field.label == field.LABEL_REPEATED  : # memeber type list of prirmitive_types
			setattr(self,f'_{field_name}' , SdkListType(typ = self.__sdk__.PROTOBUF_TYPE_TO_PYTHON_TYPE.get(field.type) ,  PARENT = self , name = field_name) )		
		if field.type != field.TYPE_MESSAGE and field.label != field.LABEL_REPEATED  : # memeber type prirmitive
			setattr(self , f'_{field_name}',field.default_value)	




	def to_proto(self) : 
		self.on_proto()
		instance = self.___naitive___()
		for field_name in self.fields() : 
			attr = getattr(self,field_name,None) # field did not intialized 
			if attr is not None : 
				attr = getattr(self,f'_{field_name}')
				field = instance.DESCRIPTOR.fields_by_name.get(field_name)
				if field.type == field.TYPE_MESSAGE and field.label != field.LABEL_REPEATED  : # memeber type message
					memeber_type  = field.message_type._concrete_class.__name__ 
					getattr(instance , field_name).CopyFrom(attr.to_proto())
				if field.type == field.TYPE_MESSAGE and field.label == field.LABEL_REPEATED  : # memeber type list of message
					[getattr(instance , field_name).append(elem.to_proto()) for elem in attr ] 
				if field.type != field.TYPE_MESSAGE and field.label == field.LABEL_REPEATED  : # memeber type list of prirmitive_types
					[getattr(instance , field_name).append(elem) for elem in attr ] 				
				if field.type != field.TYPE_MESSAGE and field.label != field.LABEL_REPEATED  : # memeber type prirmitive
					setattr(instance , field_name,attr)	
		return instance 

	def on_proto(self) : 
		return True 
		pass 

	def update(self , **kwargs) : 
		for k , v in kwargs.items() : 
			setattr(self,k , v )
	@classmethod
	def fields(cls) : 
		return {property_name for property_name, field in cls.___naitive___.DESCRIPTOR.fields_by_name.items() }

	@classmethod
	def enums(cls) : 
		return {enum_name for enum_name in cls.___naitive___.DESCRIPTOR.enum_types_by_name  }


	def __preinit__(self, **kwargs): 
		pass 

	def __postinit__(self, **kwargs): 
		pass 

	def on_field_changed(self,msg = "") : 
		pass


class DefoldSdk : 
	PROTOBUF_TYPE_TO_PYTHON_TYPE = {
		FieldDescriptor.TYPE_DOUBLE: 'float',
		FieldDescriptor.TYPE_FLOAT: 'float',
		FieldDescriptor.TYPE_INT64: 'int',
		FieldDescriptor.TYPE_UINT64: 'int',
		FieldDescriptor.TYPE_INT32: 'int',
		FieldDescriptor.TYPE_FIXED64: 'int',
		FieldDescriptor.TYPE_FIXED32: 'int',
		FieldDescriptor.TYPE_BOOL: 'bool',
		FieldDescriptor.TYPE_STRING: 'str',
		FieldDescriptor.TYPE_BYTES: 'bytes',
		FieldDescriptor.TYPE_UINT32: 'int',
		FieldDescriptor.TYPE_ENUM: 'int',
		FieldDescriptor.TYPE_SFIXED32: 'int',
		FieldDescriptor.TYPE_SFIXED64: 'int',
		FieldDescriptor.TYPE_SINT32: 'int',
		FieldDescriptor.TYPE_SINT64: 'int',
	}

	@classmethod
	def get_type_enums(cls , typename ) : 
		enums = dict()
		defold_type = getattr(Defold , typename)
		for enum_name , enum_type in defold_type.DESCRIPTOR.enum_types_by_name.items() : 
			enums[enum_name] = {
				f'{typename}.{enum_name}.{key}' : "-"
				for key , value  in enum_type.values_by_name.items()
			}
		return enums

	@classmethod
	def GenerateDocs(cls , docs = "docs/PyDefoldSdk"  ) : 
		os.makedirs(docs,exist_ok=True)
		for typ_name, typ in Defold._asdict().items() : 
			cls.get_type_enums(typ_name)
			typeinfo = cls.inspect(typ_name , silent=True)
			md_file = os.path.join(docs,f"{typ_name}.md")
			doc = jinja2.Template(doc_template).render(
					typeinfo
			)
			print(doc , file = open(md_file,"w"))

	@classmethod
	def CreateSdkType(cls , defold_type) : 
		attributes = {
			'___naitive___' : defold_type , 
			'__sdk__' : cls 
		}
		for property_name, field in defold_type.DESCRIPTOR.fields_by_name.items(): 
			def property_getter(self, name=property_name): return self.__getAttr__(name)
			def property_setter(self, value, name=property_name): self.__setAttr__(name, value)
			attributes[f'{property_name}'] = property(property_getter , property_setter)
		## nested types 
		for member in dir(defold_type) : 
			if not (member in defold_type.DESCRIPTOR.fields_by_name) : 
				if type(getattr(defold_type,member)).__name__ == 'MessageMeta' : 
					attributes[member] = cls.CreateSdkType(getattr(defold_type,member))
		## add enum related types  
		for enum_type in defold_type.DESCRIPTOR.enum_types_by_name : 
			defold_enum = getattr(defold_type ,enum_type)
			enum_dict = {
				enum_key : enum_val
				for enum_key , enum_val in defold_enum.items()
			}
			attributes[enum_type] = collections.namedtuple(enum_type , enum_dict.keys())(**enum_dict)

		sdk_type = type(defold_type.__name__, (SdkType,), attributes)
		return sdk_type

	@classmethod
	def CreateSdk(cls) : 
		result = dict()
		for typ_name, typ in Defold._asdict().items() : 
			sdktype = cls.CreateSdkType(typ)
			result[typ_name] = sdktype
		return result 


####################################################################################
#********************************** Displatched Sdk ********************************
####################################################################################
import pkgutil , inspect , sys , importlib

class Displatcher : 
	@classmethod
	def get_submodules(cls):
		try:
			module = sys.modules[__name__]
			path = module.__path__ if hasattr(module, '__path__') else None
			prefix = module.__name__ + '.'
			submodules = []
			for info in pkgutil.iter_modules(path, prefix):
				submodules.append(info[1])
			return submodules
		except ImportError:
			print(f"Module '{module_name}' not found.")
			return []


	@classmethod
	def get_dispatched_classes(cls) : 
		result = dict()
		submodules = cls.get_submodules()
		for sub_module_name in submodules : 
			sub_module = importlib.import_module(sub_module_name)
			for name, obj in inspect.getmembers(sub_module):
				if inspect.isclass(obj)  : 
					if getattr(obj,'__sdk_patch__' , False) :
						result[obj.__name__]  = obj 
		return result 

	@classmethod
	def CreateSdk(cls) : 
		result = dict()
		originalsdk =  DefoldSdk.CreateSdk()
		_classes = cls.get_dispatched_classes()
		for _name , _cls in _classes.items() : 
			if _name in originalsdk : 
				distcls = type(_name , (_cls,originalsdk.get(_name)), dict())
				originalsdk[_name] = distcls
		sdk = collections.namedtuple('sdk' , originalsdk.keys())(**originalsdk)
		return sdk  







__all__ = ['Displatcher']
	





class EmbeddedInstanceDesc : 
	__sdk_patch__ = True 
	def __preinit__(self, **kwargs): 
		self.COMPONENTS = list()

	def get_collection_parent(self,parent = None) : 
		_parent = self.PARENT 
		if _parent is None : 
			return None
		if type(_parent).__name__ == 'CollectionDesc' : 
			return _parent
		else : 
			return self.get_collection_parent(parent=_parent)

	def on_proto(self) : 
		for cmp in self.COMPONENTS : 
			if type(cmp).__name__ == "ComponentDesc" : 
				self.data += "components {\n " + f"{cmp.to_proto()}" + "}\n"
			if type(cmp).__name__ == "EmbeddedComponentDesc": 
				self.data += "embedded_components {\n " + MessageToString(cmp.to_proto(),as_one_line = False).replace('\n',"\n  ")  + "}\n"


	def addModel(self,id ,**kwargs) : 
		cmp = sdk.EmbeddedComponentDesc(id = id , type = "model" , PARENT = self , GAME = self.GAME)
		model = sdk.ModelDesc(**kwargs , PARENT = cmp , GAME = self.GAME)
		return model 
	def addCamera(self,id ,**kwargs) : 
		cmp = sdk.EmbeddedComponentDesc(id = id , type = "camera" , PARENT = self , GAME = self.GAME)
		model = sdk.CameraDesc(**kwargs , PARENT = cmp , GAME = self.GAME)
		return model 

	def addGameObject(self,**kwargs): 
		obj = self.get_collection_parent().addGameObject(**kwargs,GAME = self.GAME , PARENT = self)
		self.children.append(obj.id)
		return obj 

	def addScriptFile(self,id ,**kwargs) : 
		component = sdk.ComponentDesc(id = id ,  PARENT = self , GAME = self.GAME)
		scriptfile = sdk.LuaSourceFile(id = id , PARENT = component , GAME = self.GAME)
		self.COMPONENTS.append(component)
		return scriptfile



import os 




class MaterialDesc : 
	__sdk_patch__ = True 
	__ext__ = ".material"
	__form__ = '{name}.material'

	def __postinit__(self, **kwargs): 
		self.max_page_count = 1

	def __preinit__(self,*args,**kwargs) : 
		self.vp = ""
		self.fp = ""
		self.CONSTANT_TYPE = dict(Viewproj=getattr(self.ConstantType,"CONSTANT_TYPE_VIEWPROJ") ,
			User=getattr(self.ConstantType,"CONSTANT_TYPE_USER" ),
			World=getattr(self.ConstantType,"CONSTANT_TYPE_WORLD" ),
			Texture=getattr(self.ConstantType,"CONSTANT_TYPE_TEXTURE" ),
			View=getattr(self.ConstantType,"CONSTANT_TYPE_VIEW" ),
			Projection=getattr(self.ConstantType,"CONSTANT_TYPE_PROJECTION" ),
			Worldview=getattr(self.ConstantType,"CONSTANT_TYPE_WORLDVIEW" ),
			User_Matrix4=getattr(self.ConstantType,"CONSTANT_TYPE_USER_MATRIX4" ),
			Worldviewproj = getattr(self.ConstantType,"CONSTANT_TYPE_WORLDVIEWPROJ" ),
			Normal=getattr(self.ConstantType,"CONSTANT_TYPE_NORMAL") 
			)
		self.WRAP_MODES = dict(
			edge = getattr(self.WrapMode,"WRAP_MODE_CLAMP_TO_EDGE"), 
			repeat = getattr(self.WrapMode,"WRAP_MODE_REPEAT"),
			mirror = getattr(self.WrapMode,"WRAP_MODE_MIRRORED_REPEAT")
		)
		self.FILTER_MODE_MAG = dict(
			default = getattr(self.FilterModeMag , "FILTER_MODE_MAG_DEFAULT") , 
			linear = getattr(self.FilterModeMag , "FILTER_MODE_MAG_LINEAR"),
			nearest = getattr(self.FilterModeMag , "FILTER_MODE_MAG_NEAREST")
		)
		self.FILTER_MODE_MIN = {
			'default'  : getattr(self.FilterModeMin , "FILTER_MODE_MIN_DEFAULT"), 
			'linear'  : getattr(self.FilterModeMin , "FILTER_MODE_MIN_LINEAR"),
			'nearest' : getattr(self.FilterModeMin , "FILTER_MODE_MIN_NEAREST" ), 
			'linear-linear'  : getattr(self.FilterModeMin , "FILTER_MODE_MIN_LINEAR_MIPMAP_LINEAR"),
			'linear-nearest'  : getattr(self.FilterModeMin , "FILTER_MODE_MIN_LINEAR_MIPMAP_NEAREST"),
			'nearest-linear'  : getattr(self.FilterModeMin , "FILTER_MODE_MIN_NEAREST_MIPMAP_LINEAR"), 
			'nearest-nearest'  : getattr(self.FilterModeMin , "FILTER_MODE_MIN_NEAREST_MIPMAP_NEAREST")
		}

	def on_proto(self) : 
		if (self.name != "") and (self.GAME is not None) : 
			saving_folder = self.GAME.saving_folder(self)
			os.makedirs(saving_folder,exist_ok = True)
			print(saving_folder)
			vp_file = os.path.join(saving_folder,f'{self.name}.vp')
			fp_file = os.path.join(saving_folder,f'{self.name}.fp')
			print(self.vp , file = open(vp_file,"w"))
			print(self.fp , file = open(fp_file,"w"))
			self.fragment_program = self.GAME.get_project_path(fp_file)
			self.vertex_program = self.GAME.get_project_path(vp_file)

	def setvertex_constant(self,name,typ,value = None) :
		constant = self.Constant(name = name , type = self.CONSTANT_TYPE.get(typ) )
		self.vertex_constants.append(constant)

	def setfragment_constant(self,name,typ,value = None ) :
		constant = self.Constant(name = name , type =  self.CONSTANT_TYPE.get(typ))
		self.fragment_constants.append(constant)

	def setvertex_space(self,typ = 'local') : 
		self.vertex_space = {'world' : getattr(self.VertexSpace,'VERTEX_SPACE_WORLD') , 'local' : getattr(self.VertexSpace,'VERTEX_SPACE_LOCAL')}[typ]

	def setTexture(self,name,u = 'edge',v = 'edge' ,min = 'linear',mag ='linear') : 
		tex = self.Sampler(name = name , 
			wrap_u = self.WRAP_MODES.get(u), wrap_v = self.WRAP_MODES.get(v),
			filter_mag = self.FILTER_MODE_MAG.get(mag) , filter_min = self.FILTER_MODE_MIN.get(min)
		)
		self.samplers.append(tex)

	def export(self) : 
		self.on_proto()











class ModelDesc : 
	__mule__ = True
	def setMaterial(self,name = "model") : 
		self.materials = [sdk.Material(name = "default" , material = self.GAME.get_material_by_name(name))]

	def setMesh(self,name) : 
		self.mesh = self.GAME.get_mesh_by_name(name)

	def setTexture(self,key,name) : 
		# check key is valid  : !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
		material_proto_file = self.GAME.projectpath_2_fullpath(self.materials[0].material)
		texture_path = self.GAME.get_material_texture_by_name(name)
		texture_found  = False 
		for index , texture in enumerate(self.materials[0].textures) : 
			if texture.sampler == key : 
				self.self.materials[0].textures[index].texture = texture_path
				texture_found  = True 
		if not texture_found  : 
			self.materials[0].textures.append(
				sdk.Texture(sampler = key, texture = texture_path)
			)
		self.on_field_changed()

	def on_field_changed(self,name = None, value = None) : 
		if self.PARENT is not None : 
			self.PARENT.data = MessageToString(self.to_proto(),as_one_line = False)









class Settings : 
    SavingPolicy  = dict(
        EmbeddedInstanceDesc = "Gameobjects" , 
        Atlas = 'Atlases' , 
        MaterialDesc  = 'Materials/{name}' , 
        CollectionDesc = 'Collections'
    )
    Templates = "https://github.com/MhadhbiXissam/defold-templates"
    defaultTemplate = "default"










class CollectionDesc : 
	__sdk_patch__ = True 
	__ext__ = ".collection"
	__sdk__ = True
	
	def __preinit__(self, **kwargs): 
		self.scale_along_z = 0

	def addGameObject(self,**kwargs): 
		obj = sdk.EmbeddedInstanceDesc(**kwargs)
		obj.GAME = self.GAME 
		obj.PARENT = self 
		self.embedded_instances.append(obj)
		return obj 

